@@ -17,9 +17,11 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+from __future__ import absolute_import
 import argparse
 import locale
 import logging
+import os
 import re
 import subprocess
 import sys
@@ -30,7 +32,10 @@ try:
 except:
     pydot = None
 
+from pycflow2dot import __version__ as _VERSION
 
+
+_COLORS = ['#eecc80', '#ccee80', '#80ccee', '#eecc80', '#80eecc']
 _DOT_RESERVED = {'graph', 'strict', 'digraph', 'subgraph', 'node', 'edge'}
 logger = logging.getLogger(__name__)
 
@@ -40,27 +45,6 @@ def bytes2str(b):
     return b.decode(encoding)
 
 
-def get_max_space(lines):
-    space = 0
-    for i in range(0, len(lines)):
-        if lines[i].startswith(space * 4 * ' '):
-            i = 0
-            space += 1
-    return space
-
-
-def get_name(line):
-    name = ''
-    for i in range(0, len(line)):
-        if line[i] == ' ':
-            pass
-        elif line[i] == '(':
-            break
-        else:
-            name += line[i]
-    return name
-
-
 def call_cflow(
         c_fname, cflow,
         numbered_nesting=True,
@@ -84,55 +68,20 @@ def call_cflow(
     return cflow_data
 
 
-def cflow2dot_old(data, offset=False, filename=''):
-    color = ['#eecc80', '#ccee80', '#80ccee', '#eecc80', '#80eecc']
-    shape = ['box', 'ellipse', 'octagon', 'hexagon', 'diamond']
-    dot = (
-        'digraph G {{\n'
-        'node [peripheries=2 style="filled,rounded" '
-        'fontname="Vera Sans Mono" color="#eecc80"];\n'
-        'rankdir=LR;\n'
-        'label="{filename}"\n'
-        'main [shape=box];\n').format(
-            filename=filename)
-    lines = data.replace('\r', '').split('\n')
-    max_space = get_max_space(lines)
-    nodes = set()
-    edges = set()
-    for i in range(0, max_space):
-        for j in range(0, len(lines)):
-            if lines[j].startswith((i + 1) * 4 * ' ') \
-                    and not lines[j].startswith((i + 2) * 4 * ' '):
-                cur_node = get_name(lines[j])
-                # node already seen ?
-                if cur_node not in nodes:
-                    nodes.add(cur_node)
-                    print('New Node: ' + cur_node)
-                # predecessor \exists ?
-                try:
-                    pred_node
-                except NameError:
-                    raise Exception(
-                        'No predecessor node defined yet! Buggy...')
-                # edge already seen ?
-                cur_edge = (pred_node, cur_node)
-                if cur_edge not in edges:
-                    edges.add(cur_edge)
-                else:
-                    continue
-                dot += (('node [color="%s" shape=%s];edge [color="%s"];\n') % (
-                        color[i % 5], shape[i % 5], color[i % 5]))
-                dot += (pred_node + '->' + cur_node + '\n')
-            elif lines[j].startswith(i * 4 * ' '):
-                pred_node = get_name(lines[j])
-            else:
-                raise Exception('bug ?')
-    dot += '}\n'
-    logger.debug('dot dump str:\n\n' + dot)
-    return dot
-
-
 def cflow2nx(cflow_str, c_fname):
+    """Return graph from output of `cflow`.
+
+    @param cflow_str: output of `cflow`
+    @type cflow_str: `str`
+    @param c_fname: name of C file
+    @type c_fname: `str`
+    @return: graph of nodes named after functions,
+        with attributes:
+        - `nest_level`: distance of call from root
+        - `src_line`: source line number or
+            `-1` if function is defined in another file
+    @rtype: `networkx.DiGraph`
+    """
     lines = cflow_str.replace('\r', '').split('\n')
     g = nx.DiGraph()
     stack = dict()
@@ -197,7 +146,7 @@ def rename_if_reserved_by_dot(word):
     return word
 
 
-def dot_preamble(c_fname, for_latex):
+def dot_preamble(c_fname, for_latex, rankdir):
     c_fname = _graph_name_for_latex(c_fname, for_latex)
     d = _graph_node_defaults()
     node_defaults = ', '.join(
@@ -205,10 +154,11 @@ def dot_preamble(c_fname, for_latex):
     dot_str = (
         'digraph G {{\n'
         'node [{node_defaults}];\n'
-        'rankdir=LR;\n'
+        'rankdir={rankdir};\n'
         'label="{c_fname}"\n'
         ).format(
             node_defaults=node_defaults,
+            rankdir=rankdir,
             c_fname=c_fname)
     return dot_str
 
@@ -226,27 +176,25 @@ def _graph_name_for_latex(c_fname, for_latex):
 def _graph_node_defaults():
     """Return default properties of nodes."""
     return dict(
-        peripheries="2", style='"filled,rounded"',
-        fontname='"Vera Sans Mono"', color='"#eecc80"')
+        peripheries='1', style='"filled,rounded"',
+        fontname='"Vera Sans Mono"', fillcolor='"#ffffff"')
 
 
 def choose_node_format(node, nest_level, src_line, defined_somewhere,
                        for_latex, multi_page):
-    colors = ['#eecc80', '#ccee80', '#80ccee', '#eecc80', '#80eecc']
     shapes = ['box', 'ellipse', 'octagon', 'hexagon', 'diamond']
     sl = '\\\\'  # after fprintf \\ and after dot \, a single slash !
     # color, shape ?
     if nest_level == 0:
-        color = colors[0]
+        color = _COLORS[0]
         shape = 'box'
     else:
-        color = colors[(nest_level - 1) % 5]
+        color = _COLORS[(nest_level - 1) % 5]
         shape = shapes[nest_level % 5]
+    if src_line == -1:
+        color = None
     # fix underscores ?
-    if for_latex:
-        label = re.sub(r'_', r'\\\\_', node)
-    else:
-        label = node
+    label = _escape_underscores(node, for_latex)
     logger.debug('Label:\n\t: ' + label)
     # src line of def here ?
     if src_line != -1:
@@ -269,19 +217,33 @@ def choose_node_format(node, nest_level, src_line, defined_somewhere,
     return (label, color, shape)
 
 
+def _escape_underscores(s, for_latex):
+    """If `for_latex`, then escape `_` in `s`."""
+    if for_latex:
+        s = re.sub(r'_', r'\\\\_', s)
+    return s
+
+
 def dot_format_node(node, nest_level, src_line, defined_somewhere,
                     for_latex, multi_page):
     label, color, shape = choose_node_format(
         node, nest_level, src_line,
         defined_somewhere,
         for_latex, multi_page)
-    dot_str = (
-        '{node}[label="{label}" '
-        'color="{color}" shape={shape}];\n\n').format(
-            node=node,
-            label=label,
-            color=color,
-            shape=shape)
+    if color is None or color == '#ffffff':
+        dot_str = (
+            '{node}[label="{label}" shape={shape}];\n\n').format(
+                node=node,
+                label=label,
+                shape=shape)
+    else:
+        dot_str = (
+            '{node}[label="{label}" '
+            'fillcolor="{color}" shape={shape}];\n\n').format(
+                node=node,
+                label=label,
+                color=color,
+                shape=shape)
     return dot_str
 
 
@@ -306,8 +268,10 @@ def node_defined_in_other_src(node, other_graphs):
     return defined_somewhere
 
 
-def dump_dot_wo_pydot(graph, other_graphs, c_fname, for_latex, multi_page):
-    dot_str = dot_preamble(c_fname, for_latex)
+def dump_dot_wo_pydot(
+        graph, other_graphs, c_fname,
+        for_latex, multi_page, rankdir):
+    dot_str = dot_preamble(c_fname, for_latex, rankdir)
     # format nodes
     for node in graph:
         node_dict = graph.nodes[node]
@@ -327,14 +291,12 @@ def dump_dot_wo_pydot(graph, other_graphs, c_fname, for_latex, multi_page):
     return dot_str
 
 
-def write_dot_file(dot_str, dot_fname):
-    try:
-        dot_path = dot_fname + '.dot'
-        with open(dot_path, 'w') as fp:
-            fp.write(dot_str)
-            logger.info('Dumped dot file.')
-    except:
-        raise Exception('Failed to save dot.')
+def _dump_dot_file(dot_str, dot_fname):
+    """Dump `dot_str` to `dot` file `dot_fname`."""
+    dot_path = dot_fname + '.dot'
+    with open(dot_path, 'w') as f:
+        f.write(dot_str)
+    logger.info('Dumped dot file.')
     return dot_path
 
 
@@ -359,8 +321,15 @@ def _annotate_graph(
             node, nest_level, src_line,
             defined_somewhere,
             for_latex, multi_page)
-        g.add_node(
-            node, label=label, color=color, shape=shape)
+        if color is None or color == '#ffffff':
+            attr = dict(label=label, shape=shape)
+        else:
+            attr = dict(
+                label=label,
+                shape=shape,
+                fillcolor=color,
+                peripheries='0')
+        g.add_node(node, **attr)
     # annotate edges
     for u, v in graph.edges():
         g.add_edge(u, v)
@@ -368,32 +337,35 @@ def _annotate_graph(
 
 
 def write_graph2dot(graph, other_graphs, c_fname, img_fname,
-                    for_latex, multi_page, layout):
+                    for_latex, multi_page, layout, rankdir):
     if pydot is None:
-        print('Pydot not found. Exporting using pycflow2dot.write_dot_file().')
+        print('Pydot not found. Exporting using native exporter.')
         dot_str = dump_dot_wo_pydot(
             graph, other_graphs, c_fname,
-            for_latex=for_latex, multi_page=multi_page)
-        dot_path = write_dot_file(dot_str, img_fname)
+            for_latex=for_latex, multi_page=multi_page,
+            rankdir=rankdir)
+        dot_path = _dump_dot_file(dot_str, img_fname)
     else:
         # dump using networkx and pydot
         g = _annotate_graph(
             graph, other_graphs, c_fname, for_latex, multi_page)
-        pydot_graph = nx.drawing.nx_pydot.to_pydot(g)
-        pydot_graph.set_splines('true')
-        if layout == 'twopi':
-            pydot_graph.set_ranksep(5)
-            pydot_graph.set_root('main')
-        else:
-            pydot_graph.set_overlap(False)
-            pydot_graph.set_rankdir('LR')
-
-        dot_path = img_fname + '.dot'
-        pydot_graph.write(dot_path, format='dot')
+        dot_path = _dump_graph_to_dot(g, img_fname, layout, rankdir)
     return dot_path
 
 
-def write_graphs2dot(graphs, c_fnames, img_fname, for_latex, multi_page, layout):
+def _set_pydot_layout(pydot_graph, layout, rankdir):
+    pydot_graph.set_splines('true')
+    if layout == 'twopi':
+        pydot_graph.set_ranksep(5)
+        pydot_graph.set_root('main')
+    else:
+        pydot_graph.set_overlap(False)
+        pydot_graph.set_rankdir(rankdir)
+
+
+def write_graphs2dot(
+        graphs, c_fnames, img_fname,
+        for_latex, multi_page, layout, rankdir):
     dot_paths = list()
     for counter, (graph, c_fname) in enumerate(zip(graphs, c_fnames)):
         other_graphs = list(graphs)
@@ -401,11 +373,126 @@ def write_graphs2dot(graphs, c_fnames, img_fname, for_latex, multi_page, layout)
         cur_img_fname = img_fname + str(counter)
         dot_path = write_graph2dot(
             graph, other_graphs, c_fname, cur_img_fname,
-            for_latex, multi_page, layout)
+            for_latex, multi_page, layout, rankdir)
         dot_paths.append(dot_path)
     return dot_paths
 
 
+def _merge_graphs(graphs, c_fnames):
+    """Compose graphs from multiple C files into a single graph."""
+    g = nx.compose_all(graphs)
+    for graph, c_fname in zip(graphs, c_fnames):
+        _annotate_nodes_with_filename(g, graph, c_fname)
+    return g
+
+
+def _annotate_nodes_with_filename(g, graph, c_fname):
+    """Copy file names to node attributes of `g`.
+
+    For each node `u` of `graph`, if the node `u` is
+    defined in file `c_fname`, then annotate the node `u`
+    in graph `g` with `c_fname`.
+
+    This function adds the attribute `file_name` to
+    all nodes in `g`.
+    """
+    for u, d in graph.nodes(data=True):
+        assert u in g, (u, g.nodes())
+        src_line_no = d['src_line']
+        if src_line_no == -1:
+            continue
+        dg = g.nodes[u]
+        # assert each function is defined in
+        # at most one file
+        assert 'file_name' not in dg, dg
+        dg['file_name'] = c_fname
+
+
+def _mark_call_paths(graph, source, target):
+    if source is None or target is None:
+        return
+    source_node = rename_if_reserved_by_dot(source)
+    target_node = rename_if_reserved_by_dot(target)
+    assert source_node in graph, source_node
+    assert target_node in graph, target_node
+    for path in nx.all_simple_paths(
+            graph, source_node, target_node):
+        for u, v in zip(path[:-1], path[1:]):
+            graph.edges[u, v]['style'] = '"dashed"'
+
+
+def _format_merged_graph(graph, for_latex):
+    """Return graph with `dot` labeling."""
+    g = nx.DiGraph()
+    g.graph['node'] = _graph_node_defaults()
+    c_fnames = _collect_file_names(graph)
+    colormap = _make_colormap(c_fnames)
+    shape = '"ellipse"'
+    for u, d in graph.nodes(data=True):
+        _format_merged_node(
+            u, d, g, for_latex, shape, colormap)
+    for u, v, d in graph.edges(data=True):
+        g.add_edge(u, v, **d)
+    return g
+
+
+def _collect_file_names(graph):
+    """Return `set` of values of node attribute `file_name`."""
+    c_fnames = set()
+    for u, d in graph.nodes(data=True):
+        c_fname = d.get('file_name')
+        if c_fname is None:
+            continue
+        c_fnames.add(c_fname)
+    return c_fnames
+
+
+def _make_colormap(c_fnames):
+    """Return `dict` that maps `c_fnames` to colors."""
+    colormap = dict()
+    n = len(_COLORS)
+    for i, fname in enumerate(c_fnames):
+        colormap[fname] = _COLORS[i % n]
+    return colormap
+
+
+def _format_merged_node(u, d, g, for_latex, shape, colormap):
+    """Add node `u` to `g` forming label using `d`."""
+    src_line = d['src_line']
+    file_name = d.get('file_name')
+    node_name = _escape_underscores(u, for_latex)
+    if file_name is None:
+        assert src_line == -1, src_line
+        label = (
+            '{node_name}\\n').format(
+                node_name=node_name)
+        attr = dict(label=label, shape=shape)
+    else:
+        label = (
+            '{node_name}\\n'
+            '{file_name}\\n'
+            '{src_line}\\n').format(
+                node_name=node_name,
+                file_name=file_name,
+                src_line=src_line)
+        fillcolor = colormap[file_name]
+        attr = dict(
+            label=label,
+            shape=shape,
+            fillcolor=fillcolor,
+            peripheries='0')
+    g.add_node(u, **attr)
+
+
+def _dump_graph_to_dot(graph, img_fname, layout, rankdir):
+    """Dump `graph` to `dot` file with base `img_fname`."""
+    pydot_graph = nx.drawing.nx_pydot.to_pydot(graph)
+    _set_pydot_layout(pydot_graph, layout, rankdir)
+    dot_path = img_fname + '.dot'
+    pydot_graph.write(dot_path, format='dot')
+    return dot_path
+
+
 def check_cflow_dot_availability():
     required = ['cflow', 'dot']
     dep_paths = list()
@@ -424,8 +511,10 @@ def check_cflow_dot_availability():
 def dot2img(dot_paths, img_format, layout):
     print('This may take some time... ...')
     for dot_path in dot_paths:
-        img_fname = str(dot_path)
-        img_fname = img_fname.replace('.dot', '.' + img_format)
+        root, ext = os.path.splitext(dot_path)
+        assert ext == '.dot', ext
+        img_fname = '{root}.{ext}'.format(
+            root=root, ext=img_format)
         dot_cmd = [layout, '-T' + img_format, '-o', img_fname, dot_path]
         logger.debug(dot_cmd)
         subprocess.check_call(dot_cmd)
@@ -489,10 +578,24 @@ def parse_args():
     parser.add_argument('-r', '--reverse', default=False, action='store_true',
                         help='pass --reverse option to cflow, '
                         + 'chart callee-caller dependencies')
+    parser.add_argument('--merge', default=False, action='store_true',
+                        help='create a single graph for multiple C files.')
+    parser.add_argument(
+        '--source', action='store',
+        help=('start node for call path highlighting. '
+            'Available only with option `--merge`.'))
+    parser.add_argument(
+        '--target', action='store',
+        help=('end node for call path highlighting'
+            'Available only with option `--merge`.'))
     parser.add_argument(
         '-g', '--layout', default='dot',
         choices=['dot', 'neato', 'twopi', 'circo', 'fdp', 'sfdp'],
         help='graphviz layout algorithm.')
+    parser.add_argument(
+        '--rankdir', default='LR',
+        choices=['TB', 'LR', 'BT', 'RL'],
+        help='graph layout direction given to `dot`.')
     parser.add_argument(
         '-x', '--exclude', default='',
         help='file listing functions to ignore')
@@ -500,6 +603,10 @@ def parse_args():
         '-v', '--verbosity', default='ERROR',
         choices=['ERROR', 'WARNING', 'INFO', 'DEBUG'],
         help='logging level')
+    parser.add_argument(
+        '--version', action='version',
+        version='cflow2dot version {version}'.format(
+            version=_VERSION))
     if len(sys.argv) == 1:
         parser.print_help()
         sys.exit(1)
@@ -522,10 +629,6 @@ def rm_excluded_funcs(list_fname, graphs):
 
 def main():
     """Run cflow, parse output, produce dot and compile it into pdf | svg."""
-    copyright_msg = 'cflow2dot'
-    print(copyright_msg)
-    # input
-    cflow, dot = check_cflow_dot_availability()
     # parse arguments
     args = parse_args()
     c_fnames = args.input_filenames
@@ -535,7 +638,11 @@ def main():
     img_fname = args.output_filename
     preproc = args.preprocess
     do_rev = args.reverse
+    merge = args.merge
+    source = args.source
+    target = args.target
     layout = args.layout
+    rankdir = args.rankdir
     exclude_list_fname = args.exclude
     # configure the logger
     logger.addHandler(logging.StreamHandler())
@@ -550,6 +657,10 @@ def main():
             img_format=img_format,
             for_latex=for_latex,
             multi_page=multi_page))
+    print('cflow2dot')
+    # input
+    cflow, dot = check_cflow_dot_availability()
+    # call `cflow`
     cflow_strs = list()
     for c_fname in c_fnames:
         cur_str = call_cflow(
@@ -562,9 +673,16 @@ def main():
         cur_graph = cflow2nx(cflow_out, c_fname)
         graphs.append(cur_graph)
     rm_excluded_funcs(exclude_list_fname, graphs)
-    dot_paths = write_graphs2dot(
-        graphs, c_fnames, img_fname, for_latex,
-        multi_page, layout)
+    if merge:
+        g = _merge_graphs(graphs, c_fnames)
+        _mark_call_paths(g, source, target)
+        g = _format_merged_graph(g, for_latex)
+        dot_path = _dump_graph_to_dot(g, img_fname, layout, rankdir)
+        dot_paths = [dot_path]
+    else:
+        dot_paths = write_graphs2dot(
+            graphs, c_fnames, img_fname, for_latex,
+            multi_page, layout, rankdir)
     dot2img(dot_paths, img_format, layout)
 
 
